@@ -4,11 +4,15 @@
 #
 # incomplete, implemented on demand
 
-from mercurial import ancestor
-from mercurial import manifest
-from mercurial import context
+from mercurial import (
+    ancestor,
+    changelog,
+    manifest,
+    match as matchmod,
+    context,
+    util,
+)
 from mercurial.node import bin, hex, nullid
-from mercurial import localrepo
 
 def _maybehex(n):
     if len(n) == 20:
@@ -20,12 +24,12 @@ class overlaymanifest(object):
         self.repo = repo
         self.tree = repo.handler.git.get_object(sha)
         self._map = None
-        self._flagmap = None
+        self._flags = None
 
     def withflags(self):
         self.load()
-        return set([path for path, flag in self._flagmap.iteritems()
-                    if flag & 020100])
+        return set([path for path, flag in self._flags.iteritems()
+                    if flag != ''])
 
     def copy(self):
         return overlaymanifest(self.repo, self.tree.id)
@@ -37,8 +41,12 @@ class overlaymanifest(object):
     def iterkeys(self):
         return iter(self.keys())
 
-    def flags(self, path):
-        self.load()
+    def load(self):
+        if self._map is not None:
+            return
+
+        self._map = {}
+        self._flags = {}
 
         def hgflag(gitflag):
             if gitflag & 0100:
@@ -48,15 +56,6 @@ class overlaymanifest(object):
             else:
                 return ''
 
-        return hgflag(self._flagmap[path])
-
-    def load(self):
-        if self._map is not None:
-            return
-
-        self._map = {}
-        self._flagmap = {}
-
         def addtree(tree, dirname):
             for entry in tree.iteritems():
                 if entry.mode & 040000:
@@ -66,10 +65,21 @@ class overlaymanifest(object):
                 else:
                     path = dirname + entry.path
                     self._map[path] = bin(entry.sha)
-                    self._flagmap[path] = entry.mode
+                    self._flags[path] = hgflag(entry.mode)
 
         addtree(self.tree, '')
 
+    def matches(self, match):
+        '''generate a new manifest filtered by the match argument'''
+        if match.always():
+            return self.copy()
+
+        mf = self.copy()
+        for fn in mf.keys():
+            if not match(fn):
+                del mf[fn]
+        return mf
+
     def iteritems(self):
         self.load()
         return self._map.iteritems()
@@ -82,15 +92,95 @@ class overlaymanifest(object):
         self.load()
         return self._map[path]
 
+    def __contains__(self, path):
+        self.load()
+        return path in self._map
+
+    def get(self, path, default=None):
+        self.load()
+        return self._map.get(path, default)
+
+    def diff(self, m2, match=None, clean=False):
+        # Older mercurial clients used diff(m2, clean=False). If a caller failed
+        # to specify clean as a keyword arg, it might get passed as match here.
+        if isinstance(match, bool):
+            clean = match
+            match = None
+
+        self.load()
+        if isinstance(m2, overlaymanifest):
+            m2.load()
+
+        # below code copied from manifest.py:manifestdict.diff
+        diff = {}
+
+        try:
+            m2flagget = m2.flags
+        except AttributeError:
+            # Mercurial <= 3.3
+            m2flagget = m2._flags.get
+
+        if match is None:
+            match = matchmod.always('', '')
+        for fn, n1 in self.iteritems():
+            if not match(fn):
+                continue
+            fl1 = self._flags.get(fn, '')
+            n2 = m2.get(fn, None)
+            fl2 = m2flagget(fn, '')
+            if n2 is None:
+                fl2 = ''
+            if n1 != n2 or fl1 != fl2:
+                diff[fn] = ((n1, fl1), (n2, fl2))
+            elif clean:
+                diff[fn] = None
+
+        for fn, n2 in m2.iteritems():
+            if fn not in self:
+                if not match(fn):
+                    continue
+                fl2 = m2flagget(fn, '')
+                diff[fn] = ((None, ''), (n2, fl2))
+
+        return diff
+
     def __delitem__(self, path):
         del self._map[path]
 
+def wrapmanifestdictdiff(orig, self, m2, match=None, clean=False):
+    '''avoid calling into lazymanifest code if m2 is an overlaymanifest'''
+    # Older mercurial clients used diff(m2, clean=False). If a caller failed
+    # to specify clean as a keyword arg, it might get passed as match here.
+    if isinstance(match, bool):
+        clean = match
+        match = None
+
+    kwargs = {
+        'clean' : clean
+    }
+    # Older versions of mercurial don't support the match arg, so only add it if
+    # it exists.
+    if match is not None:
+        kwargs['match'] = match
+    if isinstance(m2, overlaymanifest):
+        diff = m2.diff(self, **kwargs)
+        # since we calculated the diff with m2 vs m1, flip it around
+        for fn in diff:
+            c1, c2 = diff[fn]
+            diff[fn] = c2, c1
+        return diff
+    else:
+        return orig(self, m2, **kwargs)
+
 class overlayfilectx(object):
     def __init__(self, repo, path, fileid=None):
-        self.repo = repo
+        self._repo = repo
         self._path = path
         self.fileid = fileid
 
+    def repo(self):
+        return self._repo
+
     # this is a hack to skip copy detection
     def ancestors(self):
         return [self, self]
@@ -108,18 +198,25 @@ class overlayfilectx(object):
         return self.fileid
 
     def data(self):
-        blob = self.repo.handler.git.get_object(_maybehex(self.fileid))
+        blob = self._repo.handler.git.get_object(_maybehex(self.fileid))
         return blob.data
 
+    def isbinary(self):
+        return util.binary(self.data())
+
 class overlaychangectx(context.changectx):
     def __init__(self, repo, sha):
-        self.repo = repo
+        # Can't store this in self._repo because the base class uses that field
+        self._hgrepo = repo
         if not isinstance(sha, basestring):
-          sha = sha.hex()
+            sha = sha.hex()
         self.commit = repo.handler.git.get_object(_maybehex(sha))
         self._overlay = getattr(repo, 'gitoverlay', repo)
         self._rev = self._overlay.rev(bin(self.commit.id))
 
+    def repo(self):
+        return self._hgrepo
+
     def node(self):
         return bin(self.commit.id)
 
@@ -145,7 +242,13 @@ class overlaychangectx(context.changectx):
         return self.commit.message
 
     def parents(self):
-        return [overlaychangectx(self.repo, sha) for sha in self.commit.parents]
+        cl = self._hgrepo.changelog
+        parents = cl.parents(cl.node(self._rev))
+        if not parents:
+            return [self._hgrepo['null']]
+        if parents[1] == nullid:
+            parents = parents[:-1]
+        return [self._hgrepo[sha] for sha in parents]
 
     def manifestnode(self):
         return bin(self.commit.tree)
@@ -177,9 +280,13 @@ class overlaychangectx(context.changectx):
         try:
             from mercurial import phases
             return phases.draft
-        except ImportError:
+        except (AttributeError, ImportError):
             return 1
 
+    def totuple(self):
+        return (self.commit.tree, self.user(), self.date(), self.files(),
+                self.description(), self.extra())
+
 class overlayrevlog(object):
     def __init__(self, repo, base):
         self.repo = repo
@@ -187,11 +294,14 @@ class overlayrevlog(object):
 
     def parents(self, n):
         gitrev = self.repo.revmap.get(n)
-        if not gitrev:
+        if gitrev is None:
             # we've reached a revision we have
             return self.base.parents(n)
         commit = self.repo.handler.git.get_object(_maybehex(n))
 
+        if not commit.parents:
+            return [nullid, nullid]
+
         def gitorhg(n):
             hn = self.repo.handler.map_hg_get(hex(n))
             if hn is not None:
@@ -211,10 +321,10 @@ class overlayrevlog(object):
         anode = self.repo.nodemap.get(a)
         bnode = self.repo.nodemap.get(b)
         if anode is None and bnode is None:
-          return self.base.ancestor(a, b)
+            return self.base.ancestor(a, b)
         ancs = ancestor.ancestors(self.parentrevs, a, b)
         if ancs:
-          return min(map(self.node, ancs))
+            return min(map(self.node, ancs))
         return nullid
 
     def parentrevs(self, rev):
@@ -229,38 +339,86 @@ class overlayrevlog(object):
     def rev(self, n):
         gitrev = self.repo.revmap.get(n)
         if gitrev is None:
-             return self.base.rev(n)
+            return self.base.rev(n)
         return gitrev
 
-    def nodesbetween(self, nodelist, revs):
-        # this is called by pre-1.9 incoming with the nodelist we returned from
-        # getremotechanges. Just return it back.
-        return [nodelist]
-
     def __len__(self):
         return len(self.repo.handler.repo) + len(self.repo.revmap)
 
-class overlaymanifestlog(overlayrevlog):
+class overlayoldmanifestlog(overlayrevlog):
     def read(self, sha):
         if sha == nullid:
             return manifest.manifestdict()
         return overlaymanifest(self.repo, sha)
 
+    def __getitem__(self, sha):
+        return overlaymanifestctx(self.repo, sha)
+
+class overlaymanifestrevlog(overlayrevlog):
+    pass
+
+class overlaymanifestctx(object):
+    def __init__(self, repo, node):
+        self._repo = repo
+        self._node = node
+
+    def read(self):
+        return overlaymanifest(self._repo, self._node)
+
+try:
+    class overlaymanifestlog(manifest.manifestlog):
+        def __init__(self, repo):
+            self._repo = repo
+
+        # Needed for 4.0, since __getitem__ did not redirect to get() in that
+        # release.
+        def __getitem__(self, node):
+            return self.get('', node)
+
+        def get(self, dir, node):
+            if dir:
+                raise RuntimeError("hggit doesn't support treemanifests")
+            if node == nullid:
+                return manifest.manifestctx()
+            return overlaymanifestctx(self._repo, node)
+except AttributeError:
+    # manifestlog did not exist prior to 4.0
+    pass
+
 class overlaychangelog(overlayrevlog):
     def read(self, sha):
         if isinstance(sha, int):
             sha = self.node(sha)
         if sha == nullid:
             return (nullid, "", (0, 0), [], "", {})
-        return overlaychangectx(self.repo, sha)
+        try:
+            return self.base.read(sha)
+        except LookupError:
+            return overlaychangectx(self.repo, sha).totuple()
 
+    def changelogrevision(self, noderev):
+        values = self.read(noderev)
+        return changelog._changelogrevision(
+            manifest=values[0],
+            user=values[1],
+            date=values[2],
+            files=values[3],
+            description=values[4],
+            extra=values[5],
+        )
 
 class overlayrepo(object):
     def __init__(self, handler, commits, refs):
         self.handler = handler
 
         self.changelog = overlaychangelog(self, handler.repo.changelog)
-        self.manifest = overlaymanifestlog(self, handler.repo.manifest)
+        if util.safehasattr(handler.repo, 'manifest'):
+            self.manifest = overlayoldmanifestlog(self, handler.repo.manifest)
+            # new as of mercurial 3.9+
+            self.manifestlog = self.manifest
+        else:
+            # no more manifest class as of 4.1
+            self.manifestlog = overlaymanifestlog(self)
 
         # for incoming -p
         self.root = handler.repo.root
@@ -275,6 +433,17 @@ class overlayrepo(object):
 
         self._makemaps(commits, refs)
 
+        try:
+            # Mercurial >= 3.3
+            from mercurial import namespaces
+            self.names = namespaces.namespaces()
+        except (AttributeError, ImportError):
+            pass
+
+    def _constructmanifest(self):
+        return overlaymanifestrevlog(self,
+                self.handler.repo._constructmanifest())
+
     def __getitem__(self, n):
         if n not in self.revmap:
             return self.handler.repo[n]
@@ -288,19 +457,19 @@ class overlayrepo(object):
         r.handler = self.handler
         r.gitoverlay = self
         try:
-          return getattr(r, method)(*args, **kwargs)
+            return getattr(r, method)(*args, **kwargs)
         finally:
-          if oldhandler is nothing:
-            del r.handler
-          else:
-            r.handler = oldhandler
-          if oldoverlay is nothing:
-            del r.gitoverlay
-          else:
-            r.gitoverlay = oldoverlay
+            if oldhandler is nothing:
+                del r.handler
+            else:
+                r.handler = oldhandler
+            if oldoverlay is nothing:
+                del r.gitoverlay
+            else:
+                r.gitoverlay = oldoverlay
 
     def status(self, *args, **kwargs):
-      return self._handlerhack('status', *args, **kwargs)
+        return self._handlerhack('status', *args, **kwargs)
 
     def node(self, n):
         """Returns an Hg or Git hash for the specified Git hash"""
@@ -320,6 +489,9 @@ class overlayrepo(object):
     def filectx(self, path, fileid=None):
         return overlayfilectx(self, path, fileid=fileid)
 
+    def unfiltered(self):
+        return self.handler.repo.unfiltered()
+
     def _makemaps(self, commits, refs):
         baserev = self.handler.repo['tip'].rev()
         self.revmap = {}
